#define PRG_NAME "WRT 4.4 (31.03.05) - Word Replacing Transformation by P.Skibinski, inikep@o2.pl"

#define CHAR_CR_LF			5
#define CHAR_NOSPACE		6	// for punctuation marks modeling
#define CHAR_ESCAPE			7	// for encode reserved chars (CHAR_ESCAPE,CHAR_FIRSTUPPER,...)
#define CHAR_FIRSTUPPER		8	// for encode lower word with first capital letter
#define CHAR_UPPERWORD		11	// for encode upper word
#define CHAR_LOWERWORD		12	// for encode lower word with few capital letter
#define NGRAM_FIRST			'A'
#define NGRAM_LAST			'Z'
#define BINARY_FIRST		128 
#define BINARY_LAST			255 


#define OPTION_TO_LOWER_AFTER_PUNCTUATION	1
#define OPTION_SPACELESS_WORDS				2
#define OPTION_CAPTIAL_CONVERSION			4
#define OPTION_SPACE_AFTER_CC_FLAG			8
#define OPTION_USE_DICTIONARY				16
#define OPTION_TRY_SHORTER_WORD				32
#define OPTION_WORD_SURROROUNDING_MODELING	64
#define OPTION_NORMAL_TEXT_FILTER			128
#define OPTION_EOL_CODING					256
#define OPTION_ADD_SYMBOLS_0_4				512
#define OPTION_ADD_SYMBOLS_14_31			1024
#define OPTION_ADD_SYMBOLS_A_Z				2048
#define OPTION_ADD_SYMBOLS_MISC				4096
#define OPTION_USE_NGRAMS					8192
#define OPTION_RECORD_INTERLEAVING			16384
#define OPTION_DNA_QUARTER_BYTE				32768

#define AUTO_SWITCH			8	// param for !OPTION_NORMAL_TEXT_FILTER
#define WORD_MIN_SIZE		1
#define FUNCTION_CHECK_ERRORS
#define WRT_HEADER "WRT4"

#define min(a,b) ((a>b)?(b):(a))
#define PUNCTUATION_CHARS(c) (c=='!' || c=='(' || c==')' || c=='-' || c=='+' || c=='[' || c==']' || c==':' || c==';' || c=='"' || c=='\'' ||  c==',' || c=='.' || c=='>' || c=='?' || c=='/')
#define TOLOWER(c) ((c>='A' && c<='Z')?(c+32):((upperSet[0][c]>0)?lowerSetRev[0][upperSet[0][c]]:c))
#define TOUPPER(c) ((c>='a' && c<='z')?(c-32):((lowerSet[0][c]>0)?upperSetRev[0][lowerSet[0][c]]:c))
#define TOUPPER_SET(c) ((c>='a' && c<='z')?(c-32):((lowerSet[usedSet][c]>0)?upperSetRev[usedSet][lowerSet[usedSet][c]]:c))
#define IF_OPTION(option) ((preprocFlag & option)!=0)
#define TURN_OFF(option) {if (preprocFlag & option) preprocFlag-=option;}
#define TURN_ON(option) {if ((preprocFlag & option)==0) preprocFlag+=option;}
#define RESET_OPTIONS (preprocFlag=0)

enum EPreprocessType { LZ77, BWT, PPM, PAQ };
enum EWordType { LOWERWORD, FIRSTUPPER, UPPERWORD };
enum EEOLType { UNDEFINED, CRLF, LF };
enum EUpperType { UFALSE, UTRUE, FORCE };
enum ESpaceType { NONE, SPACE, EOL };


#define COND_BIN_FILTER(c) (((c<32)?(c!=10 && c!=13):(0)) || (c>=BINARY_FIRST)) 



bool initOrder,forceNormalTextFilter,forceWordSurroroundModeling,forceEOLcoding;

EPreprocessType preprocType;
ESpaceType spaceBefore=NONE;
int preprocFlag=0,tryShorterBound=0;
int fftell,originalFileLen;



#define DECODE_GETC(c,file)\
{\
	c=getc(file); \
}


#define ENCODE_PUTC(c,file) \
{ \
	putc(c,file); \
}

#include <stdio.h>
#include "RangeCoder.hpp"
RangeCoder coder;
#include "Order1.hpp"
#include "Dictionary.hpp"
#include "Detect.hpp"


int bufferedChar,autoSwitch,lastEOL,lastChar,EOLcount;
EEOLType EOLType;


bool swapCase=false;

#define SWAP_CASE(c) \
{\
\
	if (swapCase)\
	{\
		if (c!=' ' && c!='\r' && c!='\n' && c!='\'' && c!='"' && c!='\t')\
			swapCase=false;\
\
		if (c>='a' && c<='z')\
		{\
			c-=32; /* toupper(c); */ \
		}\
		else\
		if (c>='A' && c<='Z')\
		{\
			c+=32; /* tolower(c); */ \
		}\
	}\
	else\
	if (c=='.' || c=='!' || c=='?')\
		swapCase=true;\
}


#define ENCODE_GETC(c,file)\
{\
	c=getc(file);\
	fftell++;\
\
	if (IF_OPTION(OPTION_TO_LOWER_AFTER_PUNCTUATION))\
		SWAP_CASE(c);\
}

#define DECODE_PUTC(c,file)\
{\
	if (IF_OPTION(OPTION_TO_LOWER_AFTER_PUNCTUATION))\
		SWAP_CASE(c);\
\
	if (fftell<originalFileLen) \
	{ \
		fftell++;\
		putc(c,file);\
	} \
}


// char WRT_get_options(FILE* file,int fileType);


// preprocess the file
void WRT_encode(FILE* file,FILE* fileout,int fileLen) //,int fileType)
{
	unsigned char s[1024];
	EWordType wordType;
	int preprocessing,c,next_c,last_c,s_size;

//	c=WRT_get_options(file,fileType);
//	ENCODE_PUTC(c,fileout);

	preprocessing=0;
	s_size=0;
	last_c=0;
	lastEOL=0;
	EOLcount=0;
	EOLType=UNDEFINED;
	wordType=LOWERWORD;


	printf("- encoding file (%d bytes)\n",fileLen);

	if (!IF_OPTION(OPTION_NORMAL_TEXT_FILTER) && !IF_OPTION(OPTION_USE_DICTIONARY))
	{
		autoSwitch=fileLen;
		preprocessing=autoSwitch;
	}
	else
		autoSwitch=AUTO_SWITCH;


	ENCODE_GETC(c,file);
	fftell=0;

	initOrder=true;
	lastEOL=0;

	while (!feof(file))
	{
		if (preprocessing>0)
		{
			if (!IF_OPTION(OPTION_NORMAL_TEXT_FILTER) && COND_BIN_FILTER(c))
			{
				preprocessing=autoSwitch;
			}
			else
			{
				preprocessing--;
				if (preprocessing==0)
					initOrder=true;
			}

			ENCODE_PUTC(c,fileout);
			ENCODE_GETC(c,file);
			continue;
		}



		if (c>='a' && c<='z' || lowerSet[usedSet][c]>0)
		{
			if (usedSet>0 && lowerSet[usedSet][c]>0)
				c=lowerSetRev[0][lowerSet[usedSet][c]];

			if (s_size==0)
			{
				wordType=LOWERWORD;
			}
			else
			{
				if (wordType==UPPERWORD)
				{
					encodeWord(fileout,s,s_size,wordType);
					if (IF_OPTION(OPTION_CAPTIAL_CONVERSION))
						ENCODE_PUTC(CHAR_LOWERWORD,fileout);

					wordType=LOWERWORD;
					s_size=1;
					s[0]=c;
					last_c=c;
					ENCODE_GETC(c,file);
					continue;
				}
			}

			s[s_size++]=c;
			if (s_size>=sizeof(s)-1)
			{
				encodeWord(fileout,s,s_size,wordType);
				s_size=0;
			}
			last_c=c;
			ENCODE_GETC(c,file);
			continue;
		}


		if (c<='Z' && c>='A' || upperSet[usedSet][c]>0)
		{
			if (usedSet>0 && upperSet[usedSet][c]>0)
				c=upperSetRev[0][upperSet[usedSet][c]];

			if (s_size==0)
			{
				wordType=FIRSTUPPER;
			}
			else
			{
				if (wordType==FIRSTUPPER)
				{
					if (s_size==1)
						wordType=UPPERWORD;
					else
					{
						encodeWord(fileout,s,s_size,wordType);

						wordType=FIRSTUPPER;
						s_size=1;
						s[0]=c;
						last_c=c;
						ENCODE_GETC(c,file);
						continue;
					}
				}
				else if (wordType==LOWERWORD)
				{
					encodeWord(fileout,s,s_size,wordType);

					wordType=FIRSTUPPER;
					s_size=1;
					s[0]=c;
					last_c=c;
					ENCODE_GETC(c,file);
					continue;
				}
			}

			s[s_size++]=c;
			if (s_size>=sizeof(s)-1)
			{
				encodeWord(fileout,s,s_size,wordType);
				s_size=0;
			}

			last_c=c;
			ENCODE_GETC(c,file);
			continue;
		}


		if (reservedSet[c])
		{
			encodeWord(fileout,s,s_size,wordType);
			s_size=0;
			ENCODE_PUTC(CHAR_ESCAPE,fileout);
			ENCODE_PUTC(c,fileout);

			if (!IF_OPTION(OPTION_NORMAL_TEXT_FILTER) && COND_BIN_FILTER(c))
				preprocessing=autoSwitch;

			last_c=c;
			ENCODE_GETC(c,file);
			continue;
		}

		encodeWord(fileout,s,s_size,wordType);

		s_size=0;

		ENCODE_GETC(next_c,file);


		if ( IF_OPTION(OPTION_WORD_SURROROUNDING_MODELING) && PUNCTUATION_CHARS(c) &&
			(last_c==10 || last_c==32 || (last_c>='a' && last_c<='z') || (last_c>='A' && last_c<='Z')) )
		{ 
			ENCODE_PUTC(32,fileout);
		} 


		if (IF_OPTION(OPTION_EOL_CODING) && (c==32 || c==10 || (c==13 && next_c==10)))
		{
			if (initOrder)
			{
				initOrder=false;
				INIT_ORDER1;
			}


			if (c==13)
			{
				if (EOLType==CRLF || EOLType==UNDEFINED)
				{
					c=next_c;
					ENCODE_GETC(next_c,file);
					lastEOL++;

					if (preprocType==LZ77)
						last_c=ContextEncodeLZ(last_c,c,next_c,fftell-lastEOL+(next_c<0?1:0));
					else
						last_c=ContextEncode(last_c,c,next_c,fftell-lastEOL+(next_c<0?1:0));


					if (EOLType==UNDEFINED && last_c!=c)
					{
						EOLType=CRLF;
						EncodeEOLformat(EOLType);
					}

					lastEOL=fftell;

					if (last_c==10)
					{
						ENCODE_PUTC(CHAR_CR_LF,fileout);
					}
					else
					{
						if ((IF_OPTION(OPTION_SPACELESS_WORDS)) && last_c==' ')
							spaceBefore=SPACE;
						else
							ENCODE_PUTC(last_c,fileout);
					}
				}
				else
					ENCODE_PUTC(c,fileout);
			}
			else
			{
				if (c==10 && EOLType==CRLF)
					ENCODE_PUTC(c,fileout)
				else
				{
					if (IF_OPTION(OPTION_EOL_CODING))
					{
						if (preprocType==LZ77)
							last_c=ContextEncodeLZ(last_c,c,next_c,fftell-lastEOL+(next_c<0?1:0));
						else	
							last_c=ContextEncode(last_c,c,next_c,fftell-lastEOL+(next_c<0?1:0));
					}
					else
						last_c=c;

					if (EOLType==UNDEFINED && last_c!=c)
					{
						EOLType=LF;
						EncodeEOLformat(EOLType);
					}

					if ((IF_OPTION(OPTION_SPACELESS_WORDS)) && last_c==' ')
						spaceBefore=SPACE;
					else
						ENCODE_PUTC(last_c,fileout);
				}

				if (c==10)
				{
					lastEOL=fftell;
				}
			}
		}
		else
		{
			if (c!=EOF)
			{
				if ((IF_OPTION(OPTION_WORD_SURROROUNDING_MODELING)) && PUNCTUATION_CHARS(c))
				{
						if (next_c>='a' && next_c<='z')
						{
							ENCODE_PUTC(c,fileout);
							ENCODE_PUTC(CHAR_LOWERWORD,fileout);
							ENCODE_PUTC(' ',fileout);
						}
						else
						{
							ENCODE_PUTC(c,fileout);
						}
				}
				else 
					if ((IF_OPTION(OPTION_SPACELESS_WORDS)) && c==' ')
						spaceBefore=SPACE;
					else
						ENCODE_PUTC(c,fileout);
			}
		}

		last_c=c;
		c=next_c;
	}

	encodeWord(fileout,s,s_size,wordType);
	s_size=0;
}



int stringCachePtr=0;
char stringCache[40];
int lastLastChar;
bool skipSpace=false;

inline void hook_putc(int c, FILE* file)
{
	if (bufferedChar<0 && c==' ') 
	{
		bufferedChar=c;
		return;
	}

	if (bufferedChar>=0)
	{
		if (IF_OPTION(OPTION_WORD_SURROROUNDING_MODELING) && PUNCTUATION_CHARS(c) &&
			(lastChar==10 || lastChar==32 || (lastChar>='a' && lastChar<='z') || (lastChar>='A' && lastChar<='Z')))
		{
			if (lastChar==32 &&	lastLastChar==-1)
				lastChar=lastLastChar;
			else
				lastChar=0;
			skipSpace=true;
		}


		if (IF_OPTION(OPTION_EOL_CODING))
		{
			if (initOrder)
			{
				initOrder=false;
				INIT_ORDER1;
			}

			if (preprocType==LZ77)
				bufferedChar=ContextDecodeLZ(lastChar,c,fftell-lastEOL);
			else
				bufferedChar=ContextDecode(lastChar,c,fftell-lastEOL);

			if (bufferedChar==10)
			{
				if (EOLType==UNDEFINED)
					EOLType=DecodeEOLformat();
				if (EOLType==CRLF)
				{
					lastChar=13;
					DECODE_PUTC(lastChar,file);
				}

				lastEOL=fftell;
			}
		}


		if (skipSpace)
			skipSpace=false;
		else
			DECODE_PUTC(bufferedChar,file);

		if (c==' ')
		{
			lastLastChar=lastChar;
			lastChar=bufferedChar;
			bufferedChar=32;
			return;
		}
		lastLastChar=-1;
		bufferedChar=-1;
	}
	
	if (c==10)
		lastEOL=fftell;


	lastChar=c;


	stringCache[stringCachePtr]=c;
	stringCachePtr++;
	if (stringCachePtr>=20)
		stringCachePtr=0;

	if (c==EOF)
		return;

	DECODE_PUTC(c,file);
	return;
}


void readEOLstream(FILE* file)
{
	unsigned int EOLstream_len=(unsigned int)(-1);
	int fileLen;

	fseek(file, -4, SEEK_END );
	fileLen=ftell(file)+4;

	fread((void*)&EOLstream_len,sizeof(EOLstream_len),1,file);

//	printf("End-of-Line (EOL) stream %d\n",EOLstream_len);

	if (EOLstream_len>0 && EOLstream_len<fileLen)
	{
		fseek(file, -((int)EOLstream_len)-4, SEEK_END );

		coder.StartDecode(file,EOLstream_len);
	}

	fseek(file, 0, SEEK_SET );
}



void writeEOLstream(FILE* fileout,int fileLen)
{
	unsigned int EOLstream_len;
	unsigned int enc;

	EOLstream_len=0;

	if (EOLcount>0)
	{
		enc=coder.FinishEncode(fileout);
		EOLstream_len=enc;
	}

	fwrite((void*)&EOLstream_len,sizeof(EOLstream_len),1,fileout);

	if (EOLcount>0)
	{
		printf("- writing End-of-Line (EOL) stream (%d bytes)\n", enc);
		printf("- encoding finished (%d->%d bytes)\n",fileLen,ftell(fileout));
	}
	else
		printf("- encoding finished (%d->%d bytes)\n",fileLen,ftell(fileout));

}



void WRT_decode(FILE* file,FILE* fileout,int decomprFileLen)
{
	bool upper=false;
	EUpperType upperWord=UFALSE;
	int i,c;
	int preprocessing=0;
	unsigned char s[1024];
	int s_size=0;

	if (IF_OPTION(OPTION_SPACELESS_WORDS))
		spaceBefore=SPACE;
	else
		spaceBefore=NONE;
	

//	fseek(file, header_len, SEEK_SET ); // skip "WRTx" header

//	fread((void*)&fileEnd,sizeof(fileEnd),1,file); // decompressedfileLen

	initOrder=true;

//	printf("- decoding file (%d bytes->%d bytes), EOL stream=%d bytes\n",fileLen,fileEnd);

//		fseek(file, header_len+4, SEEK_SET ); // skip "WRTx"+fileLen header


	printf("- decoding file (%d bytes)\n",decomprFileLen);
	
	originalFileLen=decomprFileLen;
	bufferedChar=-1;
	lastChar=0;
	fftell=0;


//	DECODE_GETC(c,file);
//	WRT_set_options(c);


	if (!IF_OPTION(OPTION_NORMAL_TEXT_FILTER) && !IF_OPTION(OPTION_USE_DICTIONARY))
	{
		autoSwitch=decomprFileLen;
		preprocessing=autoSwitch;
	}
	else
		autoSwitch=AUTO_SWITCH;

	if (IF_OPTION(OPTION_SPACELESS_WORDS))
		spaceBefore=SPACE;
	else
		spaceBefore=NONE;


	DECODE_GETC(c,file);
	lastEOL=-1;
	EOLType=UNDEFINED;
	EOLcount=0;


	while (fftell<originalFileLen)
	{
		if (preprocessing>0)
		{
			DECODE_PUTC(c,fileout);

			if (!IF_OPTION(OPTION_NORMAL_TEXT_FILTER) && COND_BIN_FILTER(c))
			{
				preprocessing=autoSwitch;
			}
			else
			{
				preprocessing--;
				if (preprocessing==0)
					initOrder=true;
			}

			DECODE_GETC(c,file);
			continue;
		}


		if (addSymbols[c] && IF_OPTION(OPTION_USE_DICTIONARY))
		{
			if (spaceBefore==SPACE)
			{
				if (upperWord==FORCE)
					upperWord=UTRUE;
				else
                    upperWord=UFALSE;
			}

			if (preprocType==LZ77)
				s_size=decodeCodeWord_LZ(file,s,c);
			else
				s_size=decodeCodeWord(file,s,c);

			if (upper)
			{
				upper=false;
				s[0]=TOUPPER(s[0]);
			}

			if (upperWord!=UFALSE)
				toUpper(s,s_size);


			if (IF_OPTION(OPTION_SPACELESS_WORDS))
			{
				if (spaceBefore==SPACE)
					hook_putc(' ',fileout); 
				else
					spaceBefore=SPACE;
			}

			for (i=0; i<s_size; i++)
			{
				ORIGINAL_CHARSET(s[i]);
				hook_putc(s[i],fileout);
			}

			DECODE_GETC(c,file);
			continue;
		}


		if (reservedSet[c])
		{
			if (c==CHAR_ESCAPE)
			{
				upper=false;
				upperWord=UFALSE;

				DECODE_GETC(c,file);
				hook_putc(c,fileout);

				if (!IF_OPTION(OPTION_NORMAL_TEXT_FILTER) && COND_BIN_FILTER(c))
					preprocessing=autoSwitch;

				DECODE_GETC(c,file);
				continue;
			}

			if (c==CHAR_NOSPACE && IF_OPTION(OPTION_SPACELESS_WORDS))
			{
				if (upperWord==FORCE)
					upperWord=UTRUE;

				DECODE_GETC(c,file);
				spaceBefore=NONE;
				continue;
			}

			if (c==CHAR_CR_LF)
			{
				hook_putc('\r',fileout);
				c='\n';
			}

			if (c==CHAR_FIRSTUPPER)
			{
				if (IF_OPTION(OPTION_SPACE_AFTER_CC_FLAG))
				{			
					DECODE_GETC(c,file); // skip space
				}
				upper=true;
				upperWord=UFALSE;
				DECODE_GETC(c,file);
				continue;
			}
			
			if (c==CHAR_UPPERWORD)
			{
				if (IF_OPTION(OPTION_SPACE_AFTER_CC_FLAG))
				{
					DECODE_GETC(c,file); // skip space
				}
				upperWord=FORCE;
				DECODE_GETC(c,file);
				continue;
			}
			
			if (c==CHAR_LOWERWORD)
			{
				upper=false;
				upperWord=UFALSE;
				DECODE_GETC(c,file);
				if (c==' ') // skip space
				{
					DECODE_GETC(c,file);
				}
				continue;
			}
		}


		if (upperWord!=UFALSE)
		{
			if (upperWord==FORCE)
				upperWord=UTRUE;

			if ((c>='a' && c<='z') || lowerSet[usedSet][c]>0)
				c=TOUPPER_SET(c);
			else
				upperWord=UFALSE;
		}
		else
		if (upper==true)
		{
			upper=false;
			c=TOUPPER_SET(c);
		}

		hook_putc(c,fileout);

		DECODE_GETC(c,file);
		continue;
	}

	hook_putc(EOF,fileout);

	printf("- decoding finished (%d bytes)\n",ftell(fileout));
}

void quarter_byte_encode(FILE* file,FILE* fileout)
{
	int c,d,flen,quart;

	fseek(file, 0, SEEK_END );
	flen=ftell(file);
	fseek(file, 0, SEEK_SET );

	quart=flen%4;
	fputc(quart,fileout);



	if (quart>0)
		quart=flen/4+1;
	else
		quart=flen/4;


	do
	{
		d=0;
		for (int j=0; j<4; j++)
		{
			c=fgetc(file);
			switch (c)
			{
				case -1:
				case 'a':
					d=0+(d<<2);
					break;
				case 'c':
					d=1+(d<<2);
					break;
				case 'g':
					d=2+(d<<2);
					break;
				case 't':
					d=3+(d<<2);
					break;
				default:
					printf("error: use WRT with -q option\n");
					return;
			}
		}
		fputc(d,fileout);
	}
	while (c!=-1);

}


#define QUARTER_DECODE(c) (c<2) ? ((c==0)?'a':'c') : ((c==2)?'g':'t')

void quarter_byte_decode(FILE* file,FILE* fileout,int header)
{
	int d,b1,b2,b3,b4,flen,quart;

	fseek(file, 0, SEEK_END );
	flen=ftell(file)-header;
	fseek(file, header, SEEK_SET );

	quart=fgetc(file);


	for (int i=2; i<=flen; i++)
	{
		d=fgetc(file);

		b4=d%4;
		d=d>>2;
		b3=d%4;
		d=d>>2;
		b2=d%4;
		d=d>>2;
		b1=d%4;

		if (i!=flen)
		{
			fputc(QUARTER_DECODE(b1),fileout);
			fputc(QUARTER_DECODE(b2),fileout);
			fputc(QUARTER_DECODE(b3),fileout);
			fputc(QUARTER_DECODE(b4),fileout);
		}
		else
		{
			switch (quart)
			{
				case 0:
					break;
				case 1:
					fputc(QUARTER_DECODE(b1),fileout);
					break;
				case 2:
					fputc(QUARTER_DECODE(b1),fileout);
					fputc(QUARTER_DECODE(b2),fileout);
					break;
				case 3:
					fputc(QUARTER_DECODE(b1),fileout);
					fputc(QUARTER_DECODE(b2),fileout);
					fputc(QUARTER_DECODE(b3),fileout);
					break;
			}			
		}
	}
}


void interleave(FILE* file,FILE* fileout,int record,int header)
{
	int i,j,flen,count;
	char* buffer;

	fseek(file, 0, SEEK_END );
	flen=ftell(file)-header;
	fseek(file, header, SEEK_SET );

	if (header>0)
		record=flen/record;


	buffer=(char*)malloc(flen);
	if (buffer==NULL)
	{
		printf("Not enough memory!\n");
		return;
	}

	fread(buffer,1,flen,file);

	count=flen/record;

	for (i=0; i<record; i++)
	{
		for (j=0; j<count; j++)
			putc(buffer[i+j*record],fileout);
	}

	j=flen%record;
	for (i=0; i<j; i++)
		putc(buffer[count*record+i],fileout);
}

int defaultSettings(int argc, char* argv[])
{
	static bool firstTime=true;

	RESET_OPTIONS;

	TURN_ON(OPTION_EOL_CODING);
	TURN_ON(OPTION_TRY_SHORTER_WORD);
	TURN_ON(OPTION_USE_DICTIONARY);
	TURN_ON(OPTION_NORMAL_TEXT_FILTER);
	TURN_ON(OPTION_CAPTIAL_CONVERSION);
	TURN_ON(OPTION_ADD_SYMBOLS_A_Z);
	TURN_ON(OPTION_RECORD_INTERLEAVING);
	TURN_ON(OPTION_DNA_QUARTER_BYTE);

	switch (preprocType)
	{
		case LZ77: // for LZ77 there are different EOLs and codeWords
			TURN_ON(OPTION_TO_LOWER_AFTER_PUNCTUATION);
			TURN_ON(OPTION_SPACELESS_WORDS);
			TURN_ON(OPTION_ADD_SYMBOLS_0_4);
			TURN_ON(OPTION_ADD_SYMBOLS_14_31);
			TURN_ON(OPTION_ADD_SYMBOLS_MISC);
			break;
		case BWT:
			TURN_ON(OPTION_SPACELESS_WORDS);
			TURN_ON(OPTION_ADD_SYMBOLS_0_4);
			TURN_ON(OPTION_ADD_SYMBOLS_14_31);
			TURN_ON(OPTION_ADD_SYMBOLS_MISC);
			break;
		case PAQ:
			tryShorterBound=4;
			TURN_OFF(OPTION_RECORD_INTERLEAVING);
			TURN_OFF(OPTION_DNA_QUARTER_BYTE);
			TURN_OFF(OPTION_ADD_SYMBOLS_A_Z);
			TURN_ON(OPTION_WORD_SURROROUNDING_MODELING);
			break;
		case PPM:
			tryShorterBound=4;
			TURN_ON(OPTION_USE_NGRAMS);
			TURN_ON(OPTION_WORD_SURROROUNDING_MODELING);
			TURN_ON(OPTION_SPACE_AFTER_CC_FLAG)
			break;
	}


	int optCount=0;
	int backArgc=argc;
	char** backArgv=argv;
	EPreprocessType newPreprocType=preprocType;

	while (argc>1 && (argv[1][0]=='-' || argv[1][0]=='+')) 
	{
		switch (argv[1][1])
		{
			case 'd':
				if (argv[1][0]=='-')
				{
					if (firstTime)
						printf("* Use dictionary option is off\n");
					TURN_OFF(OPTION_USE_DICTIONARY);
				}
				break;
			case 'e':
				if (argv[1][0]=='-')
				{
					if (firstTime)
						printf("* EOL-coding is off\n");
					TURN_OFF(OPTION_EOL_CODING);
				}
				else
				{
					if (firstTime)
						printf("* EOL-coding is on\n");
					forceEOLcoding=true;
				}
				break;
			case 't':
				if (argv[1][0]=='-')
				{
					if (firstTime)
						printf("* Try shoter word option is off\n");
					TURN_OFF(OPTION_TRY_SHORTER_WORD);
				}
				break;
			case 'n':
				if (argv[1][0]=='-')
				{
					if (firstTime)
						printf("* Use n-grams option is off\n");
					TURN_OFF(OPTION_USE_NGRAMS);
				}
				break;
			case 'p':
				if (argv[1][0]=='-')
				{
					if (firstTime)
						printf("* Punctuation marks modeling is off\n");
					TURN_OFF(OPTION_WORD_SURROROUNDING_MODELING);
				}
				else
				{
					if (firstTime)
						printf("* Punctuation marks modeling is on\n");
					forceWordSurroroundModeling=true;
				}
				break;
			case 'b':
				if (argv[1][0]=='-')
				{
					if (firstTime)
						printf("* Binary data filter is off\n");
					forceNormalTextFilter=true;
				}
				else
				{
					if (firstTime)
						printf("* Binary data filter is on\n");
					TURN_OFF(OPTION_NORMAL_TEXT_FILTER);
				}
				break;
			case 'r':
				if (argv[1][0]=='-')
				{
					if (firstTime)
						printf("* Use record interleaving is off\n");
					TURN_OFF(OPTION_RECORD_INTERLEAVING);
				}
				break;
			case 'q':
				if (argv[1][0]=='-')
				{
					if (firstTime)
						printf("* DNA quarter byte preprocessing is off\n");
					TURN_OFF(OPTION_DNA_QUARTER_BYTE);
				}
				break;
			case '0':
			case '1':
			case '2':
			case '3':
				if (firstTime)
				switch (argv[1][1])
				{
					case '0': printf("* LZ77 optimized preprocessing\n"); break;
					case '1': printf("* BWT optimized preprocessing\n"); break;
					case '2': printf("* PPM optimized preprocessing\n"); break;
					case '3': printf("* PAQ optimized preprocessing\n"); break;
				}

				newPreprocType=(EPreprocessType)(argv[1][1]-'0');
				break;
			default:
				printf("* Option %s ignored\n", argv[1]);
		}
		argc--;
		optCount++;
		argv++;
	}

	firstTime=false;

	if (newPreprocType!=preprocType)
	{
		preprocType=newPreprocType;
		return defaultSettings(backArgc,backArgv);
	}

	return optCount;
}


void main(int argc, char* argv[])
{
	FILE* file,*fileout;
	bool encoding;
	int i,c,recordLen=0,optCount,dictPathLen;
	unsigned char s[256];
	unsigned char t[256];
	unsigned char dictPath[256];
	s[0]=0;
	t[0]=0;

	printf(PRG_NAME "\n");

	forceNormalTextFilter=false;
	forceWordSurroroundModeling=false;
	forceEOLcoding=false;


	preprocType=PPM;  

	optCount=defaultSettings(argc,argv);

	if (argc-optCount<3 || strcmp(argv[optCount+1],argv[optCount+2])==0)
	{
		printf("\nusage: WRT.exe [options] <input_file> <output_file>\n");
		printf("  general options (which also set default additional options):\n");
		printf("    -0 = preprocessing optimized for further LZ77-based compresion\n");
		printf("    -1 = preprocessing optimized for further BWT-based compresion\n");
		printf("    -2 = preprocessing optimized for further PPM-based compresion (default)\n");
		printf("    -3 = preprocessing optimized for further PAQ-based compresion\n");
		printf("  additional options:\n");
		printf("    -b = Turn off binary data filter option (ignore auto detection)\n");
		printf("    +b = Turn on binary data filter option (ignore auto detection)\n");
		printf("    -d = Turn off usage of the dictionary and n-gram replacement (-n)\n");
		printf("    -e = Turn off EOL-coding (ignore auto detection)\n");
		printf("    +e = Turn on EOL-coding (ignore auto detection)\n");
		printf("    -n = Turn off n-gram replacement\n");
		printf("    -p = Turn off punctuation marks modeling (ignore auto detection)\n");
		printf("    +p = Turn on punctuation marks modeling (ignore auto detection)\n");
		printf("    -q = Turn off DNA quarter byte preprocessing (ignore auto detection)\n");
		printf("    -r = Turn off record interleaving option (ignore auto detection)\n");
		printf("    -t = Turn off \"try shorter word\" option\n");
		return;
	};


	file=fopen((const char*)argv[optCount+1],"rb");
	if (file==NULL)
	{
		printf("Can't open file %s\n",argv[optCount+1]);
		return;
	}

	char filebuf[32767];
	setvbuf(file, filebuf, _IOFBF, sizeof(filebuf));


	encoding=true;
	if (getc(file)==WRT_HEADER[0] && getc(file)==WRT_HEADER[1] && 
		getc(file)==WRT_HEADER[2] && getc(file)==WRT_HEADER[3])
	{
		encoding=false;
		i=0;
		c=getc(file);

		preprocType=(EPreprocessType)(c%4); // { LZ77, BWT, PPM, PAQ };

		if (optCount>0)
			printf("warning: decoding, so all options ignored\n");
		defaultSettings(0,NULL);


		if ((c&128)==0)
			TURN_OFF(OPTION_RECORD_INTERLEAVING)
		else
			TURN_ON(OPTION_RECORD_INTERLEAVING)

		if ((c&64)==0)
			TURN_OFF(OPTION_USE_DICTIONARY)
		else
			TURN_ON(OPTION_USE_DICTIONARY);

		if ((c&32)==0)
			TURN_OFF(OPTION_EOL_CODING)
		else
			TURN_ON(OPTION_EOL_CODING);

		if ((c&16)==0)
			TURN_OFF(OPTION_WORD_SURROROUNDING_MODELING)
		else
			TURN_ON(OPTION_WORD_SURROROUNDING_MODELING);

		if ((c&8)==0)
			TURN_OFF(OPTION_NORMAL_TEXT_FILTER)
		else
			TURN_ON(OPTION_NORMAL_TEXT_FILTER);

		if ((c&4)==0)
			TURN_OFF(OPTION_DNA_QUARTER_BYTE)
		else
			TURN_ON(OPTION_DNA_QUARTER_BYTE);


		if (IF_OPTION(OPTION_DNA_QUARTER_BYTE))
		{
			TURN_OFF(OPTION_USE_DICTIONARY);
			TURN_OFF(OPTION_RECORD_INTERLEAVING);
		}

		if (IF_OPTION(OPTION_RECORD_INTERLEAVING))
		{
			TURN_OFF(OPTION_USE_DICTIONARY);

			recordLen=256*getc(file);
			recordLen+=getc(file);
			i+=2;
		}

		if (IF_OPTION(OPTION_USE_DICTIONARY))
			usedSet=getc(file);

		if (IF_OPTION(OPTION_USE_DICTIONARY))
		{
			shortDictLen=getc(file);
			c=strlen(SHORT_DICTNAME);
			memcpy(t,SHORT_DICTNAME,c);
			for (i=c; i<shortDictLen+c; i++)
				t[i]=getc(file);
			c=strlen(DICTNAME_EXT);
			memcpy(t+i,DICTNAME_EXT,c);
			i+=c;
			t[i]=0;

			longDictLen=getc(file);
			c=strlen(DICTNAME);
			memcpy(s,DICTNAME,c);
			for (i=c; i<longDictLen+c; i++)
				s[i]=getc(file);
			c=strlen(DICTNAME_EXT);
			memcpy(s+i,DICTNAME_EXT,c);
			i+=c;
			s[i]=0;

			i=longDictLen+1+shortDictLen+1+(IF_OPTION(OPTION_USE_DICTIONARY)?1:0); // usedSet
		}
		else
		{
			longDictLen=0;
			shortDictLen=0;
		}
		i+=5; // WRT4
	}



#ifdef WIN32
	dictPathLen=getSourcePath((char*)dictPath,sizeof(dictPath));
#else
	dictPath[0]=0;
	dictPathLen=0;
#endif

	if (encoding)
	{
		printf("- encoding %s to %s\n",argv[optCount+1],argv[optCount+2]);

		readDicts(DICTNAME "*" DICTNAME_EXT,(char*)dictPath,dictPathLen);
		if (IF_OPTION(OPTION_RECORD_INTERLEAVING))
		{
			recordLen=WRT_detectFileType(file,10240,5);
			if (recordLen>0)
				TURN_OFF(OPTION_USE_DICTIONARY)
			else
				TURN_OFF(OPTION_RECORD_INTERLEAVING);
		}
		else
			WRT_detectFileType(file,10240,5);


		if (IF_OPTION(OPTION_DNA_QUARTER_BYTE))
		{
			TURN_OFF(OPTION_USE_DICTIONARY);
			TURN_OFF(OPTION_RECORD_INTERLEAVING);
		}

		if (IF_OPTION(OPTION_USE_DICTIONARY) && longDictLen>0)
			memcpy(s,langName[longDict],strlen((const char*)langName[longDict])+1);
		else
			longDictLen=0;

		if (IF_OPTION(OPTION_USE_DICTIONARY) && shortDictLen>0)
			memcpy(t,langName[shortDict],strlen((const char*)langName[shortDict])+1);
		else
			shortDictLen=0;

		freeDicts();
	}
	else
		printf("- decoding %s to %s\n",argv[optCount+1],argv[optCount+2]);


	if ((IF_OPTION(OPTION_WORD_SURROROUNDING_MODELING)) && (IF_OPTION(OPTION_SPACELESS_WORDS)))
	{
		printf("OPTION_WORD_SURROROUNDING_MODELING and OPTION_SPACELESS_WORDS collision\n");
		return;
	}

	if (!IF_OPTION(OPTION_NORMAL_TEXT_FILTER))
		TURN_OFF(OPTION_CAPTIAL_CONVERSION);


	if (dictPathLen>0)
	{
		dictPath[dictPathLen]=0;
		strcat((char*)dictPath,(char*)s);
		strcpy((char*)s,(char*)dictPath);

		dictPath[dictPathLen]=0;
		strcat((char*)dictPath,(char*)t);
		strcpy((char*)t,(char*)dictPath);
	}

	if (!initialize((longDictLen<=0)?NULL:s,(shortDictLen<=0)?NULL:t,encoding))
		return;


	fileout=fopen((const char*)argv[optCount+2],"wb");
	if (fileout==NULL)
	{
		printf("Can't open file %s\n",argv[optCount+2]);
		fclose(file);
		return;
	}

	char fileoutbuf[32767];
	setvbuf(fileout, fileoutbuf, _IOFBF, sizeof(fileoutbuf));


	if (encoding)
	{
		putc(WRT_HEADER[0],fileout);
		putc(WRT_HEADER[1],fileout);
		putc(WRT_HEADER[2],fileout);
		putc(WRT_HEADER[3],fileout);

		c=0;
		if (IF_OPTION(OPTION_RECORD_INTERLEAVING))
			c=c+128;
		if (IF_OPTION(OPTION_USE_DICTIONARY))
			c=c+64;
		if (IF_OPTION(OPTION_EOL_CODING))
			c=c+32;
		if (IF_OPTION(OPTION_WORD_SURROROUNDING_MODELING))
			c=c+16;
		if (IF_OPTION(OPTION_NORMAL_TEXT_FILTER))
			c=c+8;
		if (IF_OPTION(OPTION_DNA_QUARTER_BYTE))
			c=c+4;

		c+=preprocType;

		putc(c,fileout);


		if (IF_OPTION(OPTION_USE_DICTIONARY))
			putc(usedSet,fileout);


		if (IF_OPTION(OPTION_RECORD_INTERLEAVING))
		{
			putc(recordLen/256,fileout);
			putc(recordLen%256,fileout);
		}

		if (IF_OPTION(OPTION_USE_DICTIONARY))
		{
			putc(shortDictLen,fileout);
			c=strlen(SHORT_DICTNAME);
			for (i=c; i<shortDictLen+c; i++)
				fputc(langName[shortDict][i],fileout);

			putc(longDictLen,fileout);
			c=strlen(DICTNAME);
			for (i=c; i<longDictLen+c; i++)
				fputc(langName[longDict][i],fileout);
		}

	}


	if (encoding)
	{
		if (IF_OPTION(OPTION_DNA_QUARTER_BYTE))
			quarter_byte_encode(file,fileout);
		else
			if (IF_OPTION(OPTION_RECORD_INTERLEAVING))
				interleave(file,fileout,recordLen,0);
			else
			{
				unsigned int fileLen=flen(file);
				fwrite((void*)&fileLen,sizeof(fileLen),1,fileout);

				coder.StartEncode();
				WRT_encode(file,fileout,fileLen);
				writeEOLstream(fileout,fileLen);
			}
	}
	else
	{
		if (IF_OPTION(OPTION_DNA_QUARTER_BYTE))
			quarter_byte_decode(file,fileout,5);
		else
			if (IF_OPTION(OPTION_RECORD_INTERLEAVING))
				interleave(file,fileout,recordLen,7);
			else
			{
				readEOLstream(file);

				unsigned int fileLen;
				fseek(file, i, SEEK_SET ); // skip "WRTx" header
				fread((void*)&fileLen,sizeof(fileLen),1,file); // decompressedfileLen

	//			printf("decode fileLen=%d\n",fileLen);

				WRT_decode(file,fileout,fileLen);
			}
	}


	fclose(fileout);
	fclose(file);

	WRT_deinitialize();
}

